 /*****************************************************************************
 * This file is a mix of the original ATMEL_EB40A_Startup.s from              *
 * CrossWorks for ARM 1.5 and parts of the Ethernut project (www.ethernut.de) *
 *                                                                            *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE    *
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  *
 ******************************************************************************/

/*****************************************************************************
 * Copyright (c) 2001, 2002 Rowley Associates Limited.                       *
 *                                                                           *
 * This file may be distributed under the terms of the License Agreement     *
 * provided with this software.                                              *
 *                                                                           *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE   *
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. *
 *****************************************************************************/

/*****************************************************************************
 *                           Preprocessor Definitions
 *                           ------------------------
 *
 * STARTUP_FROM_RESET
 *
 *   If defined, the program will startup from power-on/reset. If not defined
 *   the program will just loop endlessly from power-on/reset.
 *
 *   This definition is not defined by default on this target because the
 *   debugger is unable to reset this target and maintain control of it over the
 *   JTAG interface. The advantage of doing this is that it allows the debugger
 *   to reset the CPU and run programs from a known reset CPU state on each run.
 *   It also acts as a safety net if you accidently download a program in FLASH
 *   that crashes and prevents the debugger from taking control over JTAG
 *   rendering the target unusable over JTAG. The obvious disadvantage of doing
 *   this is that your application will not startup without the debugger.
 *
 *   We advise that on this target you keep STARTUP_FROM_RESET undefined whilst
 *   you are developing and only define STARTUP_FROM_RESET when development is
 *****************************************************************************/

 /*****************************************************************************
 *                           Preprocessor Definitions
 *                           ------------------------
 *
 * VECTORED_IRQ_INTERRUPTS
 *
 *   Enable vectored IRQ interrupts. If defined, the PC register will be loaded
 *   with the contents of the AIC_IVR register on an IRQ exception.
 *
 *****************************************************************************/

/*
 * Some definitions from the ethernut project
 */
#define EBI_BASE                0xFFE00000  /* EBI base address. */
 
#define EBI_DBW                 0x00000003  /* Masks data bus width. */
#define EBI_DBW_16              0x00000001  /* 16-bit data bus width. */
#define EBI_DBW_8               0x00000002  /* 8-bit data bus width. */

#define EBI_NWS                 0x0000001C  /* Masks number of wait states. */
#define EBI_NWS_1               0x00000000  /* 1 wait state. */
#define EBI_NWS_2               0x00000004  /* 2 wait states. */
#define EBI_NWS_3               0x00000008  /* 3 wait states. */
#define EBI_NWS_4               0x0000000C  /* 4 wait states. */
#define EBI_NWS_5               0x00000010  /* 5 wait states. */
#define EBI_NWS_6               0x00000014  /* 6 wait states. */
#define EBI_NWS_7               0x00000018  /* 7 wait states. */
#define EBI_NWS_8               0x0000001C  /* 8 wait states. */

#define EBI_WSE                 0x00000020  /* Wait state enable. */

#define EBI_PAGES               0x00000180  /* Page size mask. */
#define EBI_PAGES_1M            0x00000000  /* 1 MByte page size. */
#define EBI_PAGES_4M            0x00000080  /* 4 MBytes page size. */
#define EBI_PAGES_16M           0x00000100  /* 16 MBytes page size. */
#define EBI_PAGES_64M           0x00000180  /* 64 MBytes page size. */

#define EBI_TDF                 0x00000E00  /* Masks data float output time clock cycles. */
#define EBI_TDF_0               0x00000000  /* No added cycles. */
#define EBI_TDF_1               0x00000200  /* 1 cycle. */
#define EBI_TDF_2               0x00000400  /* 2 cycles. */
#define EBI_TDF_3               0x00000600  /* 3 cycles. */
#define EBI_TDF_4               0x00000800  /* 4 cycles. */
#define EBI_TDF_5               0x00000A00  /* 5 cycles. */
#define EBI_TDF_6               0x00000C00  /* 6 cycles. */
#define EBI_TDF_7               0x00000E00  /* 7 cycles. */

#define EBI_BAT                 0x00001000  /* Byte access type */
#define EBI_BAT_BYTE_WRITE      0x00000000  /* Byte write access type */
#define EBI_BAT_BYTE_SELECT     0x00001000  /* Byte select access type */

#define EBI_CSEN                0x00002000  /* Chip select enable */

#define EBI_BA                  0xFFF00000  /* Page base address mask. */

/* Remap Control Register */
#define EBI_RCR          (EBI_BASE + 0x20)  /* Remap control register address. */
#define EBI_RCB                 0x00000001  /* Remap command. */

/* Memory Control Register */
#define EBI_MCR          (EBI_BASE + 0x24)  /* Memory control register address. */
#define EBI_ALE                 0x00000007  /* Address line enable */
#define EBI_ALE_16M             0x00000000  /* 16 Mbytes total address space. */
#define EBI_ALE_8M              0x00000004  /* 8 Mbytes total address space. */
#define EBI_ALE_4M              0x00000005  /* 4 Mbytes total address space. */
#define EBI_ALE_2M              0x00000006  /* 2 Mbytes total address space. */
#define EBI_ALE_1M              0x00000007  /* 1 Mbyte total address space. */

#define EBI_DRP                 0x00000010  /* Data read protocol mask. */
#define EBI_DRP_STANDARD        0x00000000  /* Standard read protocol. */
#define EBI_DRP_EARLY           0x00000010  /* Early read protocol. */



  .section .vectors, "ax"
  .code 32
  .align 0
  .global _vectors
  .global reset_handler

#ifdef __FLASH_BUILD
/*****************************************************************************
 * Exception Vectors (Prior to remap)                                        *
 *                                                                           *
 * After a RESET we have the following mapping:                              *
 *                                                                           *
 *     ------------------                                                    *
 *    |      SRAM        | 0x00300000 - 0x003FFFFF                           *
 *    |------------------|                                                   *
 *    |      FLASH       | 0x00000000 - 0x000FFFFF                           *
 *     ------------------                                                    *
 *                                                                           *
 * The _pre_remap_vectors are located at 0, and the CPU jumps to the         *
 * reset_handler. Here the _vectors will be copied to the SRAM and later     *
 * the memory will be remapped like:                                         *
 *                                                                           *
 *     ------------------                                                    *
 *    |      FLASH       | 0x10000000 - 0x103FFFFF                           *
 *    |------------------|                                                   *
 *    |      SRAM        | 0x00000000 - 0x0003FFFF                           *
 *     ------------------                                                    *
 *                                                                           *
 *****************************************************************************/
_pre_remap_vectors:
#ifdef STARTUP_FROM_RESET
  b reset_handler 
#else
  b .  /* Endless loop */
#endif
  b undef_handler
  b swi_handler
  b pabort_handler
  b dabort_handler
  .word 0 
  b irq_handler
  b fiq_handler
#endif

/*****************************************************************************
 * Exception Vectors (Post remap)                                            *
 *****************************************************************************/
_vectors:
  ldr pc, [pc, #reset_handler_address  - . - 8] /* reset */
  ldr pc, [pc, #undef_handler_address  - . - 8] /* undefined instruction */
  ldr pc, [pc, #swi_handler_address    - . - 8] /* swi handler */
  ldr pc, [pc, #pabort_handler_address - . - 8] /* abort prefetch */
  ldr pc, [pc, #dabort_handler_address - . - 8] /* abort data */
  nop

#ifdef VECTORED_IRQ_INTERRUPTS
  ldr pc, [pc, #-0xF20] /* Interrupt request, auto vectoring. */
  ldr pc, [pc, #-0xF20] /* Fast interrupt request, auto vectoring. */
#else
  ldr pc, [pc, #irq_handler_address - . - 8]    /* irq */
  ldr pc, [pc, #fiq_handler_address - . - 8]    /* fiq */
#endif

reset_handler_address:
  .word reset_handler
  
undef_handler_address:
  .word undef_handler
  
swi_handler_address:
  .word swi_handler
  
pabort_handler_address:
  .word pabort_handler
  
dabort_handler_address:
  .word dabort_handler
  
irq_handler_address:
  .word irq_handler
  
fiq_handler_address:
  .word fiq_handler

  .section .init, "ax"
  .code 32
  .align 0

/*****************************************************************************
 * Function    : reset_handler                                               *
 * Description : Configure memory and jump to the _start entrypoint.         *
 *****************************************************************************/
reset_handler:

  /* Configure memory */
  sub   r10, pc, #(+.-memory_configuration+8)
  ldr   r12, = copy_exception_vectors
  ldmia r10!, {r0-r9, r11}
  stmia r11!, {r0-r9}
      
  /* 
   * Jump to the default C runtime startup code. 
   * 
   * NOTE: This MUST be the last command here,
   * because the FLASH is mapped now.
   */
  mov   pc, r12

/*****************************************************************************
 * Function    : copy_exception_vectors                                      *
 * Description : Copy exception vectors into SRAM and jump                   *
 *               to the _start entrypoint.                                   *
 *****************************************************************************/
copy_exception_vectors:

#ifdef __FLASH_BUILD
  /* Copy exception vectors into SRAM */
  mov r8, #0x00000000
  ldr r9, =_vectors

  ldmia r9!, {r0-r7}
  stmia r8!, {r0-r7}
  ldmia r9!, {r0-r6}
  stmia r8!, {r0-r6}
#endif  

  ldr   r12, =_start
  mov   pc, r12

                  
/*****************************************************************************
 * Register settings for the following memory configuration:                 *
 *                                                                           *
 *     ------------------                                                    *
 *    |      FLASH       | 0x10000000 - 0x103FFFFF                           *
 *    |------------------|                                                   *
 *    |      SRAM        | 0x00000000 - 0x0003FFFF                           *
 *     ------------------                                                    *
 *                                                                           *
 *****************************************************************************/
memory_configuration:

  /* 
   * Chip select NCS0 enables 16-bit Flash memory at 0x10000000. 
   * We need 4 wait states. 
   */
  .word   0x10000000 | EBI_CSEN | EBI_PAGES_16M | EBI_WSE | EBI_NWS_4 | EBI_DBW_16
  
  /* 
   * Chip select NCS1 enables 16-bit Ethernet controller at 0x20000000.
   * We need 3 wait states. 
   */
  .word   0x20000000 | EBI_CSEN | EBI_BAT_BYTE_SELECT | EBI_PAGES_1M | EBI_WSE | EBI_NWS_3 | EBI_DBW_16
  
  /* 
   * Chip select NCS2 selects 8-bit NPL registers at 0x21000000.
   * The CPLD is quite fast and may run at one or zero wait states.
   * This pin is shared with GPIO pin 26.
   */
  .word   0x21000000 | EBI_CSEN | EBI_TDF_7 | EBI_PAGES_1M | EBI_WSE | EBI_NWS_2 | EBI_DBW_8
  
  /*
   * Chip select NCS3 is not used.
   * This pin is shared with GPIO pin 27 and available at the expansion port.
   */
  .word   0x30000000
  
  /* 
   * Chip select CS4 selects the 8-bit NPL expansion bus at 0x22000000.
   * This pin is shared with GPIO pin 31 and address bit 23.
   */
  .word   0x22000000 | EBI_CSEN | EBI_TDF_7 | EBI_PAGES_1M | EBI_WSE | EBI_NWS_8 | EBI_DBW_8
  
  /*
   * Chip select CS5 is not used.
   * This pin is shared with GPIO pin 30 and address bit 22.
   */
  .word   0x50000000
  
  /*
   * Chip select CS6 is not available, but used as address bit 21.
   */
  .word   0x60000000
  
  /*
   * Chip select CS7 is not available, but used as address bit 20.
   */
  .word   0x70000000
  
  /* 
   * EBI remap command.
   */
  .word   EBI_RCB
  
  /* 
   * Maximum address space of 4 MBytes per chip select.
   * This occupies A0 to A21.
   */
  .word   EBI_ALE_4M
  
  /* 
   * EBI base address.
   */
  .word   EBI_BASE                    


/*****************************************************************************
 *                                                                           *
 * Default exception handlers                                                *
 * These are declared weak symbols so they can be redefined in user code.    * 
 *                                                                           *
 *****************************************************************************/

undef_handler:
  b undef_handler
  
swi_handler:
  b swi_handler
  
pabort_handler:
  b pabort_handler
  
dabort_handler:
  b dabort_handler
  
irq_handler:
  b irq_handler
  
fiq_handler:
  b fiq_handler

  .weak undef_handler, swi_handler, pabort_handler, dabort_handler, irq_handler, fiq_handler

/*** EOF ***/


